/*
** main.c
**
** Project: 	exception-example
**
** Date:    	04/15/2015
** Revision:
** Author:  	DC
**
** Comments:    Demonstrate definition and usage of a simple general exception handler
**              - Light LD6 on illegal address fetch exception
**              - Light LD5 on trap (divide-by-zero) exception
**
**              Follow instructions below.
**
** Hardware Design Reference:
**  "chipKIT WF32 Rev B.0"
**
** Dependencies:
** 
**
** User Guide:
**
**
*/

/************************************************************************

SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE.
IN NO EVENT SHALL MICROCHIP TECHNOLOGY INC OR ITS LICENSORS BE LIABLE OR OBLIGATED
UNDER CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH OF WARRANTY,
OR OTHER LEGAL EQUITABLE THEORY ANY DIRECT OR INDIRECT DAMAGES OR EXPENSES
INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL, INDIRECT, PUNITIVE OR
CONSEQUENTIAL DAMAGES, LOST PROFITS OR LOST DATA, COST OF PROCUREMENT OF
SUBSTITUTE GOODS, TECHNOLOGY, SERVICES, OR ANY CLAIMS BY THIRD PARTIES
(INCLUDING BUT NOT LIMITED TO ANY DEFENSE THEREOF), OR OTHER SIMILAR COSTS.

**************************************************************************/

/** INCLUDES ******************************************************************/

#include <xc.h>             /* contains Vector Name/Number Macros */
#include <sys/attribs.h>    /* contains __ISR() Macros */

/** CONFIGURATION *************************************************************/

// Key Settings:
// OSCILLATOR: 8MHz XT Oscillator w. PLL
// SYSCLK = 80 MHz (set in config bits), PBCLK = 10 MHz (= SYSCLK/8 - set in config bits))
// JTAG PORT: Disabled
// WATCHDOG TIMER: Disabled
// DEBUG/PGM PINS: PGEC2/PGED2

// DEVCFG3
// USERID = No Setting
#pragma config FSRSSEL = PRIORITY_7     // SRS Select (SRS Priority 7)
#pragma config FMIIEN = ON              // Ethernet RMII/MII Enable (MII Enabled)
#pragma config FETHIO = ON              // Ethernet I/O Pin Select (Default Ethernet I/O)
#pragma config FUSBIDIO = ON            // USB USID Selection (Controlled by the USB Module)
#pragma config FVBUSONIO = ON           // USB VBUS ON Selection (Controlled by USB Module)

// DEVCFG2
#pragma config FPLLIDIV = DIV_2         // PLL Input Divider (2x Divider)
#pragma config FPLLMUL = MUL_20         // PLL Multiplier (20x Multiplier)
#pragma config UPLLIDIV = DIV_12        // USB PLL Input Divider (12x Divider)
#pragma config UPLLEN = OFF             // USB PLL Enable (Disabled and Bypassed)
#pragma config FPLLODIV = DIV_1         // System PLL Output Clock Divider (PLL Divide by 1)

// DEVCFG1
#pragma config FNOSC = PRIPLL           // Oscillator Selection Bits (Primary Osc (XT,HS,EC) w. PLL)
#pragma config FSOSCEN = OFF            // Secondary Oscillator Enable (Disabled)
#pragma config IESO = ON                // Internal/External Switch Over (Enabled)
#pragma config POSCMOD = XT             // Primary Oscillator Configuration (XT osc mode)
#pragma config OSCIOFNC = OFF           // CLKO Output Signal Active on the OSCO Pin (Disabled)
#pragma config FPBDIV = DIV_8           // Peripheral Clock Divisor (Pb_Clk is Sys_Clk/8)
#pragma config FCKSM = CSDCMD           // Clock Switching and Monitor Selection (Clock Switch Disable, FSCM Disabled)
#pragma config WDTPS = PS1048576        // Watchdog Timer Postscaler (1:1048576)
#pragma config FWDTEN = OFF             // Watchdog Timer Enable (WDT Disabled (SWDTEN Bit Controls))

// DEVCFG0
#pragma config DEBUG = OFF              // Background Debugger Enable (Debugger is disabled)
#pragma config ICESEL = ICS_PGx2        // ICE/ICD Comm Channel Select (ICE EMUC2/EMUD2 pins shared with PGC2/PGD2)
#pragma config PWP = OFF                // Program Flash Write Protect (Disable)
#pragma config BWP = OFF                // Boot Flash Write Protect bit (Protection Disabled)
#pragma config CP = OFF                 // Code Protect (Protection Disabled)

/** VARIABLES *****************************************************************/

int *error_1;       // pointer variable used to trigger illegal address exception
int error_2;        // variable used to trigger trap (divide-by-zero) exception

/* Exception Handler Variables */
/* Static in case exception condition would stop auto variable from being created */
static unsigned int _excep_code;
static unsigned int _excep_addr;

/** LOCAL MACROS **************************************************************/

#define SYS_CLK_FREQUENCY  (80000000ull)       // Fsys = 80 MHz
#define PB3_CLK_FREQUENCY SYS_CLK_FREQUENCY/8   // Fpb = 10 MHz

// LED_LD6 (RG6)
#define LED_LD6       	LATGbits.LATG6
#define TRIS_LED_LD6  	TRISGbits.TRISG6
#define LED_LD6_SET()   LATGSET = _LATG_LATG6_MASK;
#define LED_LD6_CLR()   LATGCLR = _LATG_LATG6_MASK;
#define LED_LD6_INV()   LATGINV = _LATG_LATG6_MASK;

// LED_LD5 (RF0)
#define LED_LD5       	LATFbits.LATF0
#define TRIS_LED_LD5  	TRISFbits.TRISF0
#define LED_LD5_SET()   LATFSET = _LATF_LATF0_MASK;
#define LED_LD5_CLR()   LATFCLR = _LATF_LATF0_MASK;
#define LED_LD5_INV()   LATFINV = _LATF_LATF0_MASK;

// Exception ExcCode Macros (see Table 50-12 in FRM Section-50 CPU - DS61192)

#define EXCCODE_ADDRESS_ERROR_STORE 5
#define EXCCODE_TRAP_EXCEPTION 13

/** LOCAL PROTOTYPES ********************************************************/

void InitializeSystem(void);            // Initialize hardware and global variables

/** main() ********************************************************************/

int main(void)
{
    error_1 = (int*)0xFFFFFFFF; // setting the pointer to point to a bad address

    InitializeSystem();

    //######################################################################
    //# Exercise:   - Exception Handling: Trigger an exception
    //#             - Uncomment Only ONE of the 2 lines below
    //#             - Debug/Run Code (will break at the while(1); in the exception handler)
    //#             - Examine 'Watches' Window for value of '_excep_code' and 'excep_addr'
    //#             - 'Watches Window' also configured to display CP0 'Status', Cause' and 'EPC' registers
    //#             - Examine LED LD6/LD5
    //#             - Open disassembly window and verify exception PC address
    //######################################################################

    *error_1 = 10;	// write to an illegal address to generate an exception
//    error_2 = 2 / 0;  // divide something by zero to generate a trap exception

    while(1);
    
} // main()

/******************************************************************************
 * Function:        void InitializeSystem(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        This routine takes care of all of the system
 *                  initialization that is required.
 *
 * Note:
 *
 *****************************************************************************/
void InitializeSystem(void)
{
    // PIC32MX CPU Speed Optimizations (Cache/Wait States/Peripheral Bus Clks)
    // On reset, and after c-startup:
    // - Prefetch Buffer is disabled,
    // - I Cache is disabled,
    // - PFM wait States set to max setting (7 = too slow!!!)
    // - Data Memory SRAM wait states set to max setting (1 = too slow!!!)

    // PBCLK - already set to SYSCLK/8 via config settings

    // Data Memory SRAM wait states: Default Setting = 1; set it to 0
    BMXCONbits.BMXWSDRM = 0;

    // Flash PM Wait States: MX Flash runs at 2 wait states @ 80 MHz
    CHECONbits.PFMWS = 2;

    // Prefetch-cache: Enable prefetch for cacheable PFM instructions
    CHECONbits.PREFEN = 1;

    // PIC32MX695-Specific
    // JTAG: Disable on PORTA
    DDPCONbits.JTAGEN = 0;

    /* Initialize LED LD6 */
    LED_LD6 = 0;
    TRIS_LED_LD6 = 0;

    /* Initialize LED LD5 */
    LED_LD5 = 0;
    TRIS_LED_LD5 = 0;


} // InitializeSystem()

/******************************************************************************
 * Function:        void _general_exception_handler(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        This function overrides the normal _weak_ _generic_exception_handler
 *                  which is defined in the XC32 User's Guide.
 *
 * Note:
 *
 *****************************************************************************/

void __attribute__((nomips16)) _general_exception_handler(void)
{
    // Mask off the ExcCode Field from the Cause Register
    // Refer to FRM Section-2 CPU - DS61113
    _excep_code=(_CP0_GET_CAUSE() & 0x0000007C) >> 2;
    _excep_addr=_CP0_GET_EPC();

    // Disable Interrupts
    _CP0_SET_STATUS(_CP0_GET_STATUS()&0xFFFFFFE); 

    // Examine _excep_code to identify the type of exception
    // Examine _excep_addr to find the address that caused the exception
    // Perform an indication of the exception
    switch(_excep_code)
    {
        case EXCCODE_ADDRESS_ERROR_STORE:
            LED_LD6_SET();
            break;
            
        case EXCCODE_TRAP_EXCEPTION:
            LED_LD5_SET();
            break;
            
        default:
            // Set both LEDs on any other exception
            LED_LD6_SET();
            LED_LD5_SET();
            break;
    }
    
    while(1);       // Set a breakpoint here
                    // Examine _excep_code to identify the type of exception
                    // Examine _excep_addr to find the address that caused the exception
                    // Example LEDs for indication

                    // In a real application, you would perform a useful indication
                    // of the exception, as well as a recovery mechanism
}